@@ -3,9 +3,9 @@
 # Picard, the next-generation MusicBrainz tagger
 #
 # Copyright (C) 2017 Sambhav Kothari
-# Copyright (C) 2017, 2019-2021 Laurent Monin
+# Copyright (C) 2017, 2019-2022 Laurent Monin
 # Copyright (C) 2018 Wieland Hoffmann
-# Copyright (C) 2018-2021 Philipp Wolfer
+# Copyright (C) 2018-2022 Philipp Wolfer
 # Copyright (C) 2020 dukeyin
 # Copyright (C) 2021 Bob Swift
 # Copyright (C) 2021 Vladislav Karbovskii
@@ -284,6 +284,19 @@ class RecordingInstrumentalTest(MBJSONTest):
         self.assertNotIn('lyricist', m)
 
 
+class MultiWorkRecordingTest(MBJSONTest):
+
+    filename = 'recording_multiple_works.json'
+
+    def test_recording(self):
+        m = Metadata()
+        t = Track('1')
+        recording_to_metadata(self.json_doc, m, t)
+        self.assertIn('instrumental', m.getall('~performance_attributes'))
+        self.assertEqual(m['language'], 'jpn; eng; zxx')
+        self.assertEqual(m['lyricist'], 'Satoru Kōsaki; Aki Hata; Minoru Shiraishi')
+
+
 class RecordingVideoTest(MBJSONTest):
 
     filename = 'recording_video.json'
@@ -389,6 +402,18 @@ class MediaTest(MBJSONTest):
         self.assertEqual(m['totaltracks'], '10')
 
 
+class MediaPregapTest(MBJSONTest):
+
+    filename = 'media_pregap.json'
+
+    def test_track(self):
+        m = Metadata()
+        medium_to_metadata(self.json_doc, m)
+        self.assertEqual(m['discnumber'], '1')
+        self.assertEqual(m['media'], 'Enhanced CD')
+        self.assertEqual(m['totaltracks'], '9')
+
+
 class NullMediaTest(MBJSONTest):
 
     filename = 'media_null.json'
@@ -445,6 +470,23 @@ class ArtistTranslationTest(MBJSONTest):
         (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
         self.assertEqual(artist_name, 'Ed Sheeran (en_CA)')
 
+    def test_locale_specific_match_first_exc(self):
+        settings = {
+            "standardize_tracks": False,
+            "standardize_artists": False,
+            "standardize_releases": False,
+            "translate_artist_names": True,
+            "translate_artist_names_script_exception": True,
+            "script_exceptions": [("LATIN", 0)],
+            "standardize_instruments": True,
+            "release_ars": True,
+            "preferred_release_countries": [],
+            "artist_locales": ['en_CA', 'en'],
+        }
+        self.set_config_values(settings)
+        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
+        self.assertEqual(artist_name, 'Ed Sheeran')
+
     def test_locale_specific_match_second(self):
         settings = {
             "standardize_tracks": False,
@@ -494,6 +536,45 @@ class ArtistTranslationTest(MBJSONTest):
         self.assertEqual(artist_name, 'Ed Sheeran')
 
 
+class ArtistTranslationArabicExceptionsTest(MBJSONTest):
+
+    filename = 'artist_arabic.json'
+
+    def test_locale_specific_match_first_exc1(self):
+        settings = {
+            "standardize_tracks": False,
+            "standardize_artists": False,
+            "standardize_releases": False,
+            "translate_artist_names": True,
+            "translate_artist_names_script_exception": True,
+            "script_exceptions": [("LATIN", 0)],
+            "standardize_instruments": True,
+            "release_ars": True,
+            "preferred_release_countries": [],
+            "artist_locales": ['en_CA', 'en'],
+        }
+        self.set_config_values(settings)
+        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
+        self.assertEqual(artist_name, 'Mohamed Mounir')
+
+    def test_locale_specific_match_first_exc2(self):
+        settings = {
+            "standardize_tracks": False,
+            "standardize_artists": False,
+            "standardize_releases": False,
+            "translate_artist_names": True,
+            "translate_artist_names_script_exception": True,
+            "script_exceptions": [("ARABIC", 0)],
+            "standardize_instruments": True,
+            "release_ars": True,
+            "preferred_release_countries": [],
+            "artist_locales": ['en_CA', 'en'],
+        }
+        self.set_config_values(settings)
+        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
+        self.assertEqual(artist_name, 'محمد منير')
+
+
 class ReleaseGroupTest(MBJSONTest):
 
     filename = 'release_group.json'
