@@ -3,16 +3,16 @@
 # Picard, the next-generation MusicBrainz tagger
 #
 # Copyright (C) 2007 Lukáš Lalinský
-# Copyright (C) 2010, 2014, 2018-2021 Philipp Wolfer
+# Copyright (C) 2010, 2014, 2018-2022 Philipp Wolfer
 # Copyright (C) 2012 Chad Wilson
 # Copyright (C) 2013 Michael Wiencek
-# Copyright (C) 2013, 2017-2020 Laurent Monin
+# Copyright (C) 2013, 2017-2021 Laurent Monin
 # Copyright (C) 2014, 2017 Sophist-UK
 # Copyright (C) 2016-2017 Sambhav Kothari
 # Copyright (C) 2017 Antonio Larrosa
 # Copyright (C) 2017-2018, 2021 Wieland Hoffmann
 # Copyright (C) 2018 virusMac
-# Copyright (C) 2020-2021 Bob Swift
+# Copyright (C) 2020-2022 Bob Swift
 # Copyright (C) 2021 Adam James
 #
 # This program is free software; you can redistribute it and/or
@@ -427,16 +427,16 @@ class ScriptParserTest(PicardTestCase):
         self.assertEqual("", self.parser.eval("$setmulti(test,%source%)", context))  # no return value
         self.assertEqual(context.getall("source"), context.getall("test"))
 
-    def test_cmd_setmulti_will_remove_empty_items(self):
+    def test_cmd_setmulti_will_keep_empty_items(self):
         context = Metadata()
         context["source"] = ["", "multi", ""]
         self.assertEqual("", self.parser.eval("$setmulti(test,%source%)", context))  # no return value
-        self.assertEqual(["multi"], context.getall("test"))
+        self.assertEqual(["", "multi", ""], context.getall("test"))
 
     def test_cmd_setmulti_custom_splitter_string(self):
         context = Metadata()
         self.assertEqual("", self.parser.eval("$setmulti(test,multi##valued##test##,##)", context))  # no return value
-        self.assertEqual(["multi", "valued", "test"], context.getall("test"))
+        self.assertEqual(["multi", "valued", "test", ""], context.getall("test"))
 
     def test_cmd_setmulti_empty_splitter_does_nothing(self):
         context = Metadata()
@@ -558,6 +558,9 @@ class ScriptParserTest(PicardTestCase):
         context["genre"] = ["Electronic", "Jungle", "Bardcore"]
         self.assertScriptResultEquals("$replacemulti(%genre%,Bardcore,Hardcore)", "Electronic; Jungle; Hardcore", context)
 
+        context["test"] = ["One", "Two", "Three"]
+        self.assertScriptResultEquals("$replacemulti(%test%,Two,)", "One; Three", context)
+
         context["test"] = ["One", "Two", "Three"]
         self.assertScriptResultEquals("$replacemulti(%test%,Four,Five)", "One; Two; Three", context)
 
@@ -700,6 +703,27 @@ class ScriptParserTest(PicardTestCase):
         context["source"] = "sourceval"
         self._eval_and_check_copymerge(context, ["targetval", "sourceval"])
 
+    def test_cmd_copymerge_empty_keepdupes(self):
+        context = Metadata()
+        context["target"] = ["tag1", "tag2", "tag1"]
+        context["source"] = ["tag2", "tag3", "tag2"]
+        self.parser.eval("$copymerge(target,source,)", context)
+        self.assertEqual(self.parser.context.getall("target"), ["tag1", "tag2", "tag3"])
+
+    def test_cmd_copymerge_keepdupes(self):
+        context = Metadata()
+        context["target"] = ["tag1", "tag2", "tag1"]
+        context["source"] = ["tag2", "tag3", "tag2"]
+        self.parser.eval("$copymerge(target,source,keep)", context)
+        self.assertEqual(self.parser.context.getall("target"), ["tag1", "tag2", "tag1", "tag2", "tag3", "tag2"])
+
+    def test_cmd_copymerge_nonlist_keepdupes(self):
+        context = Metadata()
+        context["target"] = "targetval"
+        context["source"] = "targetval"
+        self.parser.eval("$copymerge(target,source,keep)", context)
+        self.assertEqual(self.parser.context.getall("target"), ["targetval", "targetval"])
+
     def test_cmd_eq_any(self):
         self.assertScriptResultEquals("$eq_any(abc,def,ghi,jkl)", "")
         self.assertScriptResultEquals("$eq_any(abc,def,ghi,jkl,abc)", "1")
@@ -772,7 +796,7 @@ class ScriptParserTest(PicardTestCase):
     def test_default_NAT_filenaming(self):
         context = Metadata()
         context['artist'] = 'artist'
-        context['album'] = '[non-album tracks]'
+        context['album'] = '[standalone recordings]'
         context['title'] = 'title'
         result = self.parser.eval(DEFAULT_FILE_NAMING_FORMAT, context)
         self.assertEqual(result, 'artist/\n\ntitle')
@@ -926,6 +950,9 @@ class ScriptParserTest(PicardTestCase):
         context["baz"] = []
         self.assertScriptResultEquals("$lenmulti(%baz%)", "0", context)
         self.assertScriptResultEquals("$lenmulti(%baz%,:)", "0", context)
+        # Test empty multi-value elements
+        context["baz"] = ["one", "", "three"]
+        self.assertScriptResultEquals("$lenmulti(%baz%)", "3", context)
         # Test missing name
         self.assertScriptResultEquals("$lenmulti(,)", "0", context)
         self.assertScriptResultEquals("$lenmulti(,:)", "0", context)
@@ -1148,16 +1175,18 @@ class ScriptParserTest(PicardTestCase):
         self.assertScriptResultEquals("$substr(One; Two; Three,a,b)", "One; Two; Three", context)
         # # Tests with missing input
         self.assertScriptResultEquals("$substr(One; Two; Three,,4)", "One;", context)
+        self.assertScriptResultEquals("$substr(One; Two; Three,10)", "Three", context)
         self.assertScriptResultEquals("$substr(One; Two; Three,10,)", "Three", context)
+        self.assertScriptResultEquals("$substr(One; Two; Three,)", "One; Two; Three", context)
         self.assertScriptResultEquals("$substr(One; Two; Three,,)", "One; Two; Three", context)
         # Tests with missing input
-        areg = r"^\d+:\d+:\$substr: Wrong number of arguments for \$substr: Expected exactly 3, "
+        areg = r"^\d+:\d+:\$substr: Wrong number of arguments for \$substr: Expected between 2 and 3, "
         with self.assertRaisesRegex(ScriptError, areg):
             self.parser.eval("$substr()")
         with self.assertRaisesRegex(ScriptError, areg):
             self.parser.eval("$substr(abc)")
         with self.assertRaisesRegex(ScriptError, areg):
-            self.parser.eval("$substr(abc,0)")
+            self.parser.eval("$substr(abc,0,,)")
 
     def test_cmd_getmulti(self):
         context = Metadata()
@@ -1785,3 +1814,56 @@ class ScriptParserTest(PicardTestCase):
             self.parser.eval("$is_multi()")
         with self.assertRaisesRegex(ScriptError, areg):
             self.parser.eval("$is_multi(a,)")
+
+    def test_cmd_cleanmulti(self):
+        context = Metadata()
+        context["foo"] = ["", "one", "two"]
+        context["bar"] = ["one", "", "two"]
+        context["baz"] = ["one", "two", ""]
+
+        # Confirm initial values
+        self.assertScriptResultEquals("%foo%", "; one; two", context)
+        self.assertScriptResultEquals("%bar%", "one; ; two", context)
+        self.assertScriptResultEquals("%baz%", "one; two; ", context)
+        # Test cleaned values
+        self.assertScriptResultEquals("$cleanmulti(foo)%foo%", "one; two", context)
+        self.assertScriptResultEquals("$cleanmulti(bar)%bar%", "one; two", context)
+        self.assertScriptResultEquals("$cleanmulti(baz)%baz%", "one; two", context)
+
+        # Text clean with only empty string elements
+        context["foo"] = ["", "", ""]
+
+        # Confirm initial values
+        self.assertScriptResultEquals("%foo%", "; ; ", context)
+        # Test cleaned values
+        self.assertScriptResultEquals("$cleanmulti(foo)%foo%", "", context)
+
+        # Test clean with indirect argument
+        context["foo"] = ["", "one", "two"]
+        context["bar"] = "foo"
+
+        # Confirm initial values
+        self.assertScriptResultEquals("%foo%", "; one; two", context)
+        # Test cleaned values
+        self.assertScriptResultEquals("$cleanmulti(%bar%)%foo%", "one; two", context)
+
+        # Test clean with non-multi argument
+        context["foo"] = "one"
+        context["bar"] = "one; ; two"
+        context["baz"] = ""
+
+        # Confirm initial values
+        self.assertScriptResultEquals("%foo%", "one", context)
+        self.assertScriptResultEquals("%bar%", "one; ; two", context)
+        self.assertScriptResultEquals("%baz%", "", context)
+        # Test cleaned values
+        self.assertScriptResultEquals("$cleanmulti(foo)%foo%", "one", context)
+        self.assertScriptResultEquals("$cleanmulti(bar)%bar%", "one; ; two", context)
+        self.assertScriptResultEquals("$cleanmulti(baz)%baz%", "", context)
+
+        # Tests with invalid number of arguments
+        areg = r"^\d+:\d+:\$cleanmulti: Wrong number of arguments for \$cleanmulti: Expected exactly 1, "
+        with self.assertRaisesRegex(ScriptError, areg):
+            self.parser.eval("$cleanmulti()")
+        with self.assertRaisesRegex(ScriptError, areg):
+            self.parser.eval("$cleanmulti(foo,)")
