@@ -4,7 +4,7 @@
 #
 # Copyright (C) 2006-2007 Lukáš Lalinský
 # Copyright (C) 2010 fatih
-# Copyright (C) 2010-2011, 2014, 2018-2021 Philipp Wolfer
+# Copyright (C) 2010-2011, 2014, 2018-2022 Philipp Wolfer
 # Copyright (C) 2012, 2014, 2018 Wieland Hoffmann
 # Copyright (C) 2013 Ionuț Ciocîrlan
 # Copyright (C) 2013-2014, 2018-2021 Laurent Monin
@@ -37,30 +37,41 @@ import re
 import subprocess  # nosec: B404
 from tempfile import NamedTemporaryFile
 import unittest
-from unittest.mock import Mock
+from unittest.mock import (
+    Mock,
+    patch,
+)
 
 from test.picardtestcase import PicardTestCase
 
 from picard import util
+from picard.const import MUSICBRAINZ_SERVERS
 from picard.const.sys import (
     IS_MACOS,
     IS_WIN,
 )
 from picard.util import (
+    IgnoreUpdatesContext,
     album_artist_from_path,
+    any_exception_isinstance,
+    build_qurl,
     extract_year_from_date,
     find_best_match,
     is_absolute_path,
+    iter_exception_chain,
     iter_files_from_objects,
     iter_unique,
     limited_join,
     make_filename_from_title,
+    normpath,
     pattern_as_regex,
     sort_by_similarity,
+    system_supports_long_paths,
     tracknum_and_title_from_filename,
     tracknum_from_filename,
     uniqify,
     wildcards_to_regex_pattern,
+    win_prefix_longpath,
 )
 
 
@@ -405,16 +416,13 @@ class SortBySimilarity(PicardTestCase):
             SimMatchTest(similarity=0.75, name='c'),
         ]
 
-    def candidates(self):
-        yield from self.test_values
-
     def test_sort_by_similarity(self):
-        results = [result.name for result in sort_by_similarity(self.candidates)]
+        results = [result.name for result in sort_by_similarity(self.test_values)]
         self.assertEqual(results, ['b', 'c', 'd', 'a'])
 
     def test_findbestmatch(self):
         no_match = SimMatchTest(similarity=-1, name='no_match')
-        best_match = find_best_match(self.candidates, no_match)
+        best_match = find_best_match(self.test_values, no_match)
 
         self.assertEqual(best_match.result.name, 'b')
         self.assertEqual(best_match.similarity, 0.75)
@@ -423,7 +431,7 @@ class SortBySimilarity(PicardTestCase):
         self.test_values = []
 
         no_match = SimMatchTest(similarity=-1, name='no_match')
-        best_match = find_best_match(self.candidates, no_match)
+        best_match = find_best_match(self.test_values, no_match)
 
         self.assertEqual(best_match.result.name, 'no_match')
         self.assertEqual(best_match.similarity, -1)
@@ -515,6 +523,9 @@ class TracknumFromFilenameTest(PicardTestCase):
             (242, 'track-242.mp3'),
             (242, 'track nr 242.mp3'),
             (242, 'track_242.mp3'),
+            (3, 'track003.mp3'),
+            (40, 'Track40.mp3'),
+            (None, 'UB40.mp3'),
             (1, 'artist song 2004 track01 xxxx.ogg'),
             (1, 'artist song 2004 track-no-01 xxxx.ogg'),
             (1, 'artist song 2004 track-no_01 xxxx.ogg'),
@@ -568,11 +579,17 @@ class TracknumAndTitleFromFilenameTest(PicardTestCase):
         tests = (
             ((None, 'Foo'), 'Foo.mp3'),
             (('1', 'Track 0001'), 'Track 0001.mp3'),
+            (('42', 'Track-42'), 'Track-42.mp3'),
             (('99', 'Foo'), '99 Foo.mp3'),
             (('42', 'Foo'), '0000042 Foo.mp3'),
             (('2', 'Foo'), '0000002 Foo.mp3'),
             ((None, '20000 Feet'), '20000 Feet.mp3'),
             ((None, '20,000 Feet'), '20,000 Feet.mp3'),
+            ((None, 'Venus (Original 12" version)'), 'Venus (Original 12" version).mp3'),
+            ((None, 'Vanity 6'), 'Vanity 6.mp3'),
+            ((None, 'UB40 - Red Red Wine'), 'UB40 - Red Red Wine.mp3'),
+            ((None, 'Red Red Wine - UB40'), 'Red Red Wine - UB40.mp3'),
+            ((None, 'Symphony no. 5 in D minor'), 'Symphony no. 5 in D minor.mp3'),
         )
         for expected, filename in tests:
             result = tracknum_and_title_from_filename(filename)
@@ -671,3 +688,177 @@ class WildcardsToRegexPatternTest(PicardTestCase):
         regex = wildcards_to_regex_pattern(pattern)
         self.assertEqual(re.escape(pattern), regex)
         re.compile(regex)
+
+
+class BuildQUrlTest(PicardTestCase):
+
+    def test_path_and_querystring(self):
+        query = {'foo': 'x', 'bar': 'y'}
+        self.assertEqual('http://example.com/', build_qurl('example.com', path='/').toDisplayString())
+        self.assertEqual('http://example.com/foo/bar', build_qurl('example.com', path='/foo/bar').toDisplayString())
+        self.assertEqual('http://example.com/foo/bar?foo=x&bar=y', build_qurl('example.com', path='/foo/bar', queryargs=query).toDisplayString())
+        self.assertEqual('http://example.com?foo=x&bar=y', build_qurl('example.com', queryargs=query).toDisplayString())
+
+    def test_standard_ports(self):
+        self.assertEqual('http://example.com', build_qurl('example.com').toDisplayString())
+        self.assertEqual('http://example.com', build_qurl('example.com', port=80).toDisplayString())
+        self.assertEqual('https://example.com', build_qurl('example.com', port=443).toDisplayString())
+
+    def test_custom_port(self):
+        self.assertEqual('http://example.com:8080', build_qurl('example.com', port=8080).toDisplayString())
+        self.assertEqual('http://example.com:8080/', build_qurl('example.com', port=8080, path="/").toDisplayString())
+        self.assertEqual('http://example.com:8080?foo=x', build_qurl('example.com', port=8080, queryargs={'foo': 'x'}).toDisplayString())
+
+    def test_mb_server(self):
+        for host in MUSICBRAINZ_SERVERS:
+            expected = 'https://' + host
+            self.assertEqual(expected, build_qurl(host, port=80).toDisplayString())
+            self.assertEqual(expected, build_qurl(host, port=443).toDisplayString())
+            self.assertEqual(expected, build_qurl(host, port=8080).toDisplayString())
+
+
+class NormpathTest(PicardTestCase):
+
+    @unittest.skipIf(IS_WIN, "non-windows test")
+    def test_normpath(self):
+        self.assertEqual('/foo/bar', normpath('/foo//bar'))
+        self.assertEqual('/bar', normpath('/foo/../bar'))
+
+    @unittest.skipUnless(IS_WIN, "windows test")
+    def test_normpath_windows(self):
+        self.assertEqual('C:\\Foo\\Bar.baz', normpath('C:/Foo/Bar.baz'))
+        self.assertEqual('C:\\Bar.baz', normpath('C:/Foo/../Bar.baz'))
+
+    @unittest.skipUnless(IS_WIN, "windows test")
+    @patch.object(util, 'system_supports_long_paths')
+    def test_normpath_windows_longpath(self, mock_system_supports_long_paths):
+        mock_system_supports_long_paths.return_value = False
+        path = 'C:\\foo\\' + (252 * 'a')
+        self.assertEqual(path, normpath(path))
+        path += 'a'
+        self.assertEqual('\\\\?\\' + path, normpath(path))
+
+
+class WinPrefixLongpathTest(PicardTestCase):
+
+    def test_win_prefix_longpath_is_long(self):
+        path = 'C:\\foo\\' + (253 * 'a')
+        self.assertEqual('\\\\?\\' + path, win_prefix_longpath(path))
+
+    def test_win_prefix_longpath_is_short(self):
+        path = 'C:\\foo\\' + (252 * 'a')
+        self.assertEqual(path, win_prefix_longpath(path))
+
+
+class SystemSupportsLongPathsTest(PicardTestCase):
+
+    def setUp(self):
+        super().setUp()
+        try:
+            del system_supports_long_paths._supported
+        except AttributeError:
+            pass
+
+    def test_system_supports_long_paths_returns_bool(self):
+        result = system_supports_long_paths()
+        self.assertTrue(isinstance(result, bool))
+
+    @unittest.skipIf(IS_WIN, "non-windows test")
+    def test_system_supports_long_paths(self):
+        self.assertTrue(system_supports_long_paths())
+
+    @unittest.skipUnless(IS_WIN, "windows test")
+    @patch('winreg.OpenKey')
+    @patch('winreg.QueryValueEx')
+    def test_system_supports_long_paths_windows_unsupported(self, mock_query_value, mock_open_key):
+        mock_query_value.return_value = [0]
+        self.assertFalse(system_supports_long_paths())
+        mock_open_key.assert_called_once()
+        mock_query_value.assert_called_once()
+
+    @unittest.skipUnless(IS_WIN, "windows test")
+    @patch('winreg.OpenKey')
+    @patch('winreg.QueryValueEx')
+    def test_system_supports_long_paths_windows_supported(self, mock_query_value, mock_open_key):
+        mock_query_value.return_value = [1]
+        self.assertTrue(system_supports_long_paths())
+        mock_open_key.assert_called_once()
+        mock_query_value.assert_called_once()
+
+    @unittest.skipUnless(IS_WIN, "windows test")
+    @patch('winreg.OpenKey')
+    @patch('winreg.QueryValueEx')
+    def test_system_supports_long_paths_cache(self, mock_query_value, mock_open_key):
+        mock_query_value.return_value = [1]
+        self.assertTrue(system_supports_long_paths())
+        self.assertTrue(system_supports_long_paths._supported)
+        mock_query_value.return_value = [0]
+        self.assertTrue(system_supports_long_paths())
+        mock_open_key.assert_called_once()
+        mock_query_value.assert_called_once()
+
+
+class IterExceptionChainTest(PicardTestCase):
+
+    def test_iter_exception_chain(self):
+        e1 = Mock(name='e1')
+        e2 = Mock(name='e2')
+        e3 = Mock(name='e3')
+        e4 = Mock(name='e4')
+        e5 = Mock(name='e5')
+        e1.__context__ = e2
+        e2.__context__ = e3
+        e2.__cause__ = e4
+        e1.__cause__ = e5
+        self.assertEqual([e1, e2, e3, e4, e5], list(iter_exception_chain(e1)))
+
+
+class AnyExceptionIsinstanceTest(PicardTestCase):
+
+    def test_any_exception_isinstance_itself(self):
+        ex = RuntimeError()
+        self.assertTrue(any_exception_isinstance(ex, RuntimeError))
+
+    def test_any_exception_isinstance_context(self):
+        ex = Mock()
+        self.assertFalse(any_exception_isinstance(ex, RuntimeError))
+        ex.__context__ = RuntimeError()
+        self.assertTrue(any_exception_isinstance(ex, RuntimeError))
+
+    def test_any_exception_isinstance_cause(self):
+        ex = Mock()
+        self.assertFalse(any_exception_isinstance(ex, RuntimeError))
+        ex.__cause__ = RuntimeError()
+        self.assertTrue(any_exception_isinstance(ex, RuntimeError))
+
+    def test_any_exception_isinstance_nested(self):
+        ex = Mock()
+        self.assertFalse(any_exception_isinstance(ex, RuntimeError))
+        ex.__cause__ = Mock()
+        ex.__cause__.__context__ = RuntimeError()
+        self.assertTrue(any_exception_isinstance(ex, RuntimeError))
+
+
+class IgnoreUpdatesContextTest(PicardTestCase):
+
+    def test_enter_exit(self):
+        context = IgnoreUpdatesContext()
+        self.assertFalse(context)
+        with context:
+            self.assertTrue(context)
+        self.assertFalse(context)
+
+    def test_run_onexit(self):
+        onexit = Mock()
+        context = IgnoreUpdatesContext(onexit=onexit)
+        with context:
+            onexit.assert_not_called()
+        onexit.assert_called_once_with()
+
+    def test_nested_with(self):
+        context = IgnoreUpdatesContext()
+        with context:
+            with context:
+                self.assertTrue(context)
+            self.assertTrue(context)
+        self.assertFalse(context)
